#pragma once
#ifndef FEATURE_GENERATOR_PLUGIN_BASE_H
#define FEATURE_GENERATOR_PLUGIN_BASE_H

#include <absl/container/flat_hash_map.h>
#include <absl/strings/string_view.h>
#include <absl/types/optional.h>

#include <map>
#include <stdexcept>
#include <unordered_map>
#include <utility>
#include <vector>

#include "fsmap.h"
#include "integral_types.h"

namespace fg {

using absl::optional;
using std::string;
using std::vector;

template <typename T>
using List = std::vector<T>;
template <typename K, typename V>
using Map = absl::flat_hash_map<K, V>;
template <typename K, typename V>
using MapArray = std::vector<std::pair<K, V>>;
using Matrix = std::vector<std::vector<float>>;
using MatrixL = std::vector<std::vector<int64>>;
using MatrixS = std::vector<std::vector<string>>;
template <typename K, typename V>
using FSMap = featurestore::type::fs_map<K, V>;

using Field = absl::variant<
    optional<string>, optional<int32>, optional<int64>, optional<float>,
    optional<double>, optional<absl::string_view>,

    List<string>, List<int32>, List<int64>, List<float>, List<double>,
    List<absl::string_view>,

    Map<string, string>, Map<string, int32>, Map<string, int64>,
    Map<string, float>, Map<string, double>, Map<string, absl::string_view>,

    Map<absl::string_view, absl::string_view>, Map<absl::string_view, int32>,
    Map<absl::string_view, int64>, Map<absl::string_view, float>,
    Map<absl::string_view, double>, Map<absl::string_view, string>,

    Map<int32, string>, Map<int32, int32>, Map<int32, int64>, Map<int32, float>,
    Map<int32, double>, Map<int32, absl::string_view>,

    Map<int64, string>, Map<int64, float>, Map<int64, double>,
    Map<int64, int32>, Map<int64, int64>, Map<int64, absl::string_view>,

    FSMap<absl::string_view, absl::string_view>,
    FSMap<absl::string_view, int32>, FSMap<absl::string_view, int64>,
    FSMap<absl::string_view, float>, FSMap<absl::string_view, double>,

    FSMap<int32, int32>, FSMap<int32, int64>, FSMap<int32, float>,
    FSMap<int32, double>, FSMap<int32, absl::string_view>,

    FSMap<int64, float>, FSMap<int64, double>, FSMap<int64, int32>,
    FSMap<int64, int64>, FSMap<int64, absl::string_view>,

    MapArray<string, string>, MapArray<string, int32>, MapArray<string, int64>,
    MapArray<string, float>, MapArray<string, double>,

    MapArray<int32, string>, MapArray<int32, float>, MapArray<int32, double>,
    MapArray<int32, int32>, MapArray<int32, int64>,

    MapArray<int64, string>, MapArray<int64, float>, MapArray<int64, double>,
    MapArray<int64, int32>, MapArray<int64, int64>, Matrix, MatrixL, MatrixS>;

// represents a COLUMN of the feature table
using VariantVector = absl::variant<
    vector<optional<string>>, vector<optional<int32>>, vector<optional<int64>>,
    vector<optional<float>>, vector<optional<double>>,
    vector<optional<absl::string_view>>,

    vector<List<string>>, vector<List<int32>>, vector<List<int64>>,
    vector<List<float>>, vector<List<double>>, vector<List<absl::string_view>>,

    vector<Map<string, string>>, vector<Map<string, int32>>,
    vector<Map<string, int64>>, vector<Map<string, float>>,
    vector<Map<string, double>>, vector<Map<string, absl::string_view>>,

    vector<Map<absl::string_view, absl::string_view>>,
    vector<Map<absl::string_view, int32>>,
    vector<Map<absl::string_view, int64>>,
    vector<Map<absl::string_view, float>>,
    vector<Map<absl::string_view, double>>,

    vector<Map<int32, string>>, vector<Map<int32, int32>>,
    vector<Map<int32, int64>>, vector<Map<int32, float>>,
    vector<Map<int32, double>>, vector<Map<int32, absl::string_view>>,

    vector<Map<int64, string>>, vector<Map<int64, float>>,
    vector<Map<int64, double>>, vector<Map<int64, int32>>,
    vector<Map<int64, int64>>, vector<Map<int64, absl::string_view>>,

    vector<FSMap<absl::string_view, absl::string_view>>,
    vector<FSMap<absl::string_view, int32>>,
    vector<FSMap<absl::string_view, int64>>,
    vector<FSMap<absl::string_view, float>>,
    vector<FSMap<absl::string_view, double>>,

    vector<FSMap<int32, int32>>, vector<FSMap<int32, int64>>,
    vector<FSMap<int32, float>>, vector<FSMap<int32, double>>,
    vector<FSMap<int32, absl::string_view>>,

    vector<FSMap<int64, float>>, vector<FSMap<int64, double>>,
    vector<FSMap<int64, int32>>, vector<FSMap<int64, int64>>,
    vector<FSMap<int64, absl::string_view>>,

    vector<MapArray<string, string>>, vector<MapArray<string, int32>>,
    vector<MapArray<string, int64>>, vector<MapArray<string, float>>,
    vector<MapArray<string, double>>,

    vector<MapArray<int32, string>>, vector<MapArray<int32, float>>,
    vector<MapArray<int32, double>>, vector<MapArray<int32, int32>>,
    vector<MapArray<int32, int64>>,

    vector<MapArray<int64, string>>, vector<MapArray<int64, float>>,
    vector<MapArray<int64, double>>, vector<MapArray<int64, int32>>,
    vector<MapArray<int64, int64>>, vector<Matrix>, vector<MatrixL>,
    vector<MatrixS>>;

/**
 * @brief 自定义特征算子的公共基类
 *
 * 框架会检测子类有没有override批量接口`BatchProcess`方法，如果有实现则会调用该方法完成特征变换；
 * 否则，框架根据`value_type`的配置从一下`ProcessWith*`方法中选择一个执行，用户必须实现其中一个对应类型的接口
 */
class IFeatureOP {
 public:
  class NotOverriddenException : public std::exception {
   public:
    NotOverriddenException(const std::string& msg) : msg_(msg) {}
    const char* what() const noexcept override {
      if (msg_.empty()) {
        return "unimplemented method called";
      }
      std::string msg = "unimplemented method called: " + msg_;
      return msg.c_str();
    }

   private:
    const std::string msg_;
  };

  virtual ~IFeatureOP() = default;

  /**
   * @brief 初始化方法
   * @param feature_config is a json string,
   * @return 如果为0，则表示模型加载成功，否则表示模型加载失败。
   */
  virtual int Initialize(const string& feature_config) = 0;

  /**
   * @brief 特征变换，输出为string类型
   * @param inputs 表示一条记录，可以有多个字段（field）
   * @param outputs 特征变换的输出
   * @return 状态码，如果为0表示执行成功
   */
  virtual int ProcessWithStrOutputs(const vector<Field>& inputs,
                                    vector<string>& outputs) {
    throw NotOverriddenException("ProcessWithStrOutputs");
  }

  /**
   * @brief 特征变换，输出为int32类型
   * @param inputs 表示一条记录，可以有多个字段（field）
   * @param outputs 特征变换的输出
   * @return 状态码，如果为0表示执行成功
   */
  virtual int ProcessWithInt32Outputs(const vector<Field>& inputs,
                                      vector<int32>& outputs) {
    throw NotOverriddenException("ProcessWithInt32Outputs");
  }

  /**
   * @brief 特征变换，输出为int64类型
   * @param inputs 表示一条记录，可以有多个字段（field）
   * @param outputs 特征变换的输出
   * @return 状态码，如果为0表示执行成功
   */
  virtual int ProcessWithInt64Outputs(const vector<Field>& inputs,
                                      vector<int64>& outputs) {
    throw NotOverriddenException("ProcessWithInt64Outputs");
  }

  /**
   * @brief 特征变换，输出为float类型
   * @param inputs 表示一条记录，可以有多个字段（field）
   * @param outputs 特征变换的输出
   * @return 状态码，如果为0表示执行成功
   */
  virtual int ProcessWithFloatOutputs(const vector<Field>& inputs,
                                      vector<float>& outputs) {
    throw NotOverriddenException("ProcessWithFloatOutputs");
  }

  /**
   * @brief 特征变换，输出为double类型
   * @param inputs 表示一条记录，可以有多个字段（field）
   * @param outputs 特征变换的输出
   * @return 状态码，如果为0表示执行成功
   */
  virtual int ProcessWithDoubleOutputs(const vector<Field>& inputs,
                                       vector<double>& outputs) {
    throw NotOverriddenException("ProcessWithDoubleOutputs");
  }

  /**
   * @brief 可选，处理多个records的批量接口
   *
   * @param inputs 输入column的vector，VariantVector表示一个特征column
   * @param outputs
   * 输出，变换后的特征；支持输出复杂类型，可作为其他的特征变换的输入
   * @return 状态码，如果为0表示执行成功
   */
  virtual int BatchProcess(const vector<VariantVector>& inputs,
                           VariantVector& outputs) {
    throw NotOverriddenException("BatchProcess");
  }
};

using CreatePluginFunc = IFeatureOP* (*)();

class PluginRegistry {
 public:
  static void registerPlugin(const std::string& name, CreatePluginFunc func);
  static CreatePluginFunc getCreateFunc(const std::string& name);

  static std::map<std::string, CreatePluginFunc>& getRegistry();
};

}  // namespace fg

#define REGISTER_PLUGIN(PluginName, PluginClass)                           \
  extern "C" fg::IFeatureOP* create##PluginClass() {                       \
    return new fg::PluginClass();                                          \
  }                                                                        \
  struct PluginClass##Registrar {                                          \
    PluginClass##Registrar() {                                             \
      fg::PluginRegistry::registerPlugin(PluginName, create##PluginClass); \
    }                                                                      \
  };                                                                       \
  static PluginClass##Registrar PluginClass##_registrar;

#endif  // FEATURE_GENERATOR_PLUGIN_BASE_H