#pragma once
#include <absl/hash/hash.h>
#include <absl/strings/string_view.h>
#include <absl/types/optional.h>

#include <algorithm>
#include <cstdint>
#include <functional>
#include <stdexcept>
#include <string>
#include <type_traits>
#include <vector>
#include "integral_types.h"

namespace featurestore {
namespace type {

template <typename KeyType, typename ValueType>
class fs_map {
 public:
  // Constructor for numeric values
  fs_map(const int64_t* keys, size_t length, const ValueType* values)
      : _keys(keys), _length(length), _values(values) {}

  // Constructor for absl::string_view
  fs_map(const int64_t* keys, size_t length, const uint8_t* data,
         const uint32_t* offsets)
      : _keys(keys), _length(length), _data(data), _offsets(offsets) {}
  
  // Default constructor
  fs_map() = default;
  virtual ~fs_map() = default;

  static fg::int64 getHash(const fg::int32& key);
  static fg::int64 getHash(const fg::int64& key);
  static fg::int64 getHash(const std::string& key);
  static fg::int64 getHash(const absl::string_view& key);
  static fg::int64 getHash(const char* key);

  size_t size() const { return _length; }

  int64_t getKey(const size_t& index) const {
    if (index < _length) {
      return _keys[index];
    }
    throw std::out_of_range("Key index out of range");
  }

  absl::optional<ValueType> getValue(const size_t& index) const {
    if (index < _length) {
      return getValueAt(index);
    }
    throw std::out_of_range("Key index out of range");
  }

  size_t count(const int64_t& hashKey) const {
    if (_length == 0) {
      return 0;
    }
    auto it = std::lower_bound(_keys, _keys + _length, hashKey);
    if (it != _keys + _length && *it == hashKey) {
      return 1;
    }
    return 0;
  }

  absl::optional<ValueType> operator[](const int64_t& hashKey) const {
    if (_length == 0) {
      return absl::nullopt;
    }
    auto it = std::lower_bound(_keys, _keys + _length, hashKey);
    if (it != _keys + _length && *it == hashKey) {
      size_t index = it - _keys;
      return getValueAt(index);
    }
    return absl::nullopt;
  }

  ValueType at(const int64_t& hashKey) const {
    if (_length == 0) {
      throw std::out_of_range("Key not found");
    }
    auto it = std::lower_bound(_keys, _keys + _length, hashKey);
    if (it != _keys + _length && *it == hashKey) {
      size_t index = it - _keys;
      return getValueAt(index);
    }
    throw std::out_of_range("Key not found");
  }

 private:
  ValueType getValueAt(size_t index) const {
    if constexpr (std::is_same<ValueType, absl::string_view>::value) {
      uint32_t offset = _offsets[index];
      uint32_t next_offset = _offsets[index + 1];
      uint32_t length = next_offset - offset;
      return absl::string_view(reinterpret_cast<const char*>(_data + offset),
                               length);
    } else {
      return _values[index];
    }
  }

  // key
  const int64_t* _keys;
  size_t _length;

  // value
  union {
    const ValueType* _values;  // For numeric values
    struct {
      const uint8_t* _data;
      const uint32_t* _offsets;
    };
  };
};

// Specializations for supported KeyType when KeyType is int32_t or int64_t
#define DEFINE_HASH_SPECIALIZATION_FOR_INT_KEY(KeyType, ValueType)             \
  template <>                                                                  \
  inline fg::int64 fs_map<KeyType, ValueType>::getHash(const fg::int32& key) {     \
    return static_cast<size_t>(key);                                          \
  }                                                                            \
  template <>                                                                  \
  inline fg::int64 fs_map<KeyType, ValueType>::getHash(const fg::int64& key) {     \
    return key;                                                                \
  }                                                                            \
  template <>                                                                  \
  inline fg::int64 fs_map<KeyType, ValueType>::getHash(const std::string& key) { \
    throw std::invalid_argument("Key should be int32_t or size_t");           \
  }                                                                            \
  template <>                                                                  \
  inline fg::int64 fs_map<KeyType, ValueType>::getHash(                          \
      const absl::string_view& key) {                                          \
    throw std::invalid_argument("Key should be int32_t or size_t");           \
  }                                                                            \
  template <>                                                                  \
  inline fg::int64 fs_map<KeyType, ValueType>::getHash(const char* key) {        \
    throw std::invalid_argument("Key should be int32_t or size_t");           \
  }

#define DEFINE_HASH_SPECIALIZATION_FOR_STRING_VIEW_KEY(ValueType) \
  template <>                                                     \
  inline fg::int64 fs_map<absl::string_view, ValueType>::getHash(   \
      const fg::int32& key) {                                       \
    return absl::Hash<std::string>{}(std::to_string(key));        \
  }                                                               \
  template <>                                                     \
  inline fg::int64 fs_map<absl::string_view, ValueType>::getHash(   \
      const fg::int64& key) {                                       \
    return absl::Hash<std::string>{}(std::to_string(key));        \
  }                                                               \
  template <>                                                     \
  inline fg::int64 fs_map<absl::string_view, ValueType>::getHash(   \
      const std::string& key) {                                   \
    return absl::Hash<std::string>{}(key);                        \
  }                                                               \
  template <>                                                     \
  inline fg::int64 fs_map<absl::string_view, ValueType>::getHash(   \
      const absl::string_view& key) {                             \
    return absl::Hash<absl::string_view>{}(key);                  \
  }                                                               \
  template <>                                                     \
  inline fg::int64 fs_map<absl::string_view, ValueType>::getHash(   \
      const char* key) {                                          \
    return absl::Hash<absl::string_view>{}(key);                  \
  }

// 实例化模板，并应用特化宏
#define APPLY_HASH_SPECIALIZATION(ValueType)                  \
  template class fs_map<fg::int32, ValueType>;                  \
  template class fs_map<fg::int64, ValueType>;                  \
  template class fs_map<absl::string_view, ValueType>;        \
  DEFINE_HASH_SPECIALIZATION_FOR_INT_KEY(fg::int32, ValueType); \
  DEFINE_HASH_SPECIALIZATION_FOR_INT_KEY(fg::int64, ValueType); \
  DEFINE_HASH_SPECIALIZATION_FOR_STRING_VIEW_KEY(ValueType)

APPLY_HASH_SPECIALIZATION(fg::int32)
APPLY_HASH_SPECIALIZATION(fg::int64)
APPLY_HASH_SPECIALIZATION(float)
APPLY_HASH_SPECIALIZATION(double)
APPLY_HASH_SPECIALIZATION(absl::string_view)

#undef DEFINE_HASH_SPECIALIZATION_FOR_INT_KEY
#undef DEFINE_HASH_SPECIALIZATION_FOR_STRING_VIEW_KEY
#undef APPLY_HASH_SPECIALIZATION

}  // namespace type
}  // namespace featurestore
