#include "regex_replace.h"

#include <nlohmann/json.hpp>

#include "log.h"

namespace fg {

int RegexReplace::Initialize(const string& feature_config) {
  nlohmann::json cfg;
  try {
    cfg = nlohmann::json::parse(feature_config);
  } catch (nlohmann::json::parse_error& ex) {
    LOG(ERROR) << "parse error at byte " << ex.byte;
    LOG(ERROR) << "config: " << feature_config;
    throw std::runtime_error("parse RegexReplace config failed");
  }

  feature_name_ = cfg.at("feature_name");
  default_value_ = cfg.value("default_value", "");
  string patten = cfg.at("regex_patten");
  if (cfg.value("icase", false)) {
    patten_.assign(patten, std::regex::optimize | std::regex::icase);
  } else {
    patten_.assign(patten, std::regex::optimize);
  }
  replacement_ = cfg.value("replacement", "");
  return 0;
}

// 定义 overloaded 类
template <class... Ts>
struct overloaded : Ts... {
  using Ts::operator()...;
};
// 类模板参数推导指引（C++17）
template <class... Ts>
overloaded(Ts...) -> overloaded<Ts...>;

int RegexReplace::ProcessWithStrOutputs(const vector<Field>& inputs,
                                        vector<string>& outputs) {
  outputs.clear();
  if (inputs.empty()) {
    outputs.push_back(0);
    return -1;  // invalid inputs
  }
  absl::visit(
      overloaded{[&](const optional<string>& s1) {
                   if (default_value_.empty() &&
                       (!s1.has_value() || s1.value().empty())) {
                     return;
                   }
                   string result = std::regex_replace(s1.value_or(""), patten_,
                                                      replacement_);
                   outputs.emplace_back(result.empty() ? default_value_
                                                       : std::move(result));
                 },
                 [&](const optional<absl::string_view>& s1) {
                   if (default_value_.empty() &&
                       (!s1.has_value() || s1.value().empty())) {
                     return;
                   }
                   string result = std::regex_replace(string(s1.value_or("")),
                                                      patten_, replacement_);
                   outputs.emplace_back(result.empty() ? default_value_
                                                       : std::move(result));
                 },
                 [&](const List<absl::string_view>& strs) {
                   if (default_value_.empty() && strs.empty()) {
                     return;
                   }
                   for (auto& str : strs) {
                     string result =
                         std::regex_replace(string(str), patten_, replacement_);
                     outputs.emplace_back(result.empty() ? default_value_
                                                         : std::move(result));
                   }
                 },
                 [&](const List<string>& strs) {
                   if (default_value_.empty() && strs.empty()) {
                     return;
                   }
                   for (auto& str : strs) {
                     string result =
                         std::regex_replace(str, patten_, replacement_);
                     outputs.emplace_back(result.empty() ? default_value_
                                                         : std::move(result));
                   }
                 },
                 [this](auto x) {
                   ERROR_EXIT(feature_name_,
                              "unsupported input type: ", typeid(x).name());
                 }},
      inputs.at(0));
  return 0;
}

int RegexReplace::ProcessWithFloatOutputs(const vector<Field>& inputs,
                                          vector<float>& outputs) {
  vector<string> values;
  int status = ProcessWithStrOutputs(inputs, values);
  if (0 != status) {
    return status;
  }
  outputs.clear();
  std::transform(values.begin(), values.end(), std::back_inserter(outputs),
                 [](string& value) {
                   float v;
                   absl::SimpleAtof(value, &v);
                   return v;
                 });
  return 0;
}

int RegexReplace::ProcessWithDoubleOutputs(const vector<Field>& inputs,
                                           vector<double>& outputs) {
  vector<string> values;
  int status = ProcessWithStrOutputs(inputs, values);
  if (0 != status) {
    return status;
  }
  outputs.clear();
  std::transform(values.begin(), values.end(), std::back_inserter(outputs),
                 [](string& value) {
                   double v;
                   absl::SimpleAtod(value, &v);
                   return v;
                 });
  return 0;
}

int RegexReplace::ProcessWithInt32Outputs(const vector<Field>& inputs,
                                          vector<int32>& outputs) {
  vector<string> values;
  int status = ProcessWithStrOutputs(inputs, values);
  if (0 != status) {
    return status;
  }
  outputs.clear();
  std::transform(values.begin(), values.end(), std::back_inserter(outputs),
                 [](string& value) {
                   int32 v;
                   absl::SimpleAtoi(value, &v);
                   return v;
                 });
  return 0;
}

int RegexReplace::ProcessWithInt64Outputs(const vector<Field>& inputs,
                                          vector<int64>& outputs) {
  vector<string> values;
  int status = ProcessWithStrOutputs(inputs, values);
  if (0 != status) {
    return status;
  }
  outputs.clear();
  std::transform(values.begin(), values.end(), std::back_inserter(outputs),
                 [](string& value) {
                   int64 v;
                   absl::SimpleAtoi(value, &v);
                   return v;
                 });
  return 0;
}

}  // namespace fg